import os
from typing import Any, Dict, Callable, Union, List, Tuple

import numpy
import numpy as np

from sohstationviewer.model.mseed_data.record_reader_helper import Unpacker
from sohstationviewer.model.reftek_data.reftek_reader.packet import \
    eh_et_payload_end_in_packet
from sohstationviewer.model.reftek_data.reftek_reader.packet_readers import (
    read_dt_packet, read_eh_et_packet, read_soh_packet,
)
from sohstationviewer.model.reftek_data.reftek_reader.packets import (
    DTPacket, EHETPacket, SOHPacket,
)
from sohstationviewer.model.reftek_data.reftek_reader.header import \
    get_rt130_packet_header


def packet_reader_placeholder(*args: Any, **kwargs: Any) -> Tuple[Any, Any]:
    """
    Placeholder function to be used in place of an RT130 packet reader
    function. This function immediately returns None.
    """
    return None, None


def read_rt130_file(file_name: str, unpacker: Unpacker
                    ) -> List[Union[EHETPacket, DTPacket, SOHPacket]]:
    """
    Read an RT130 file and stores the data in a list of RT130 packets.
    :param file_name: the name of the file to read.
    :param unpacker: the decoder used to decode the data.
    :return: a list of processed RT130 packets.
    """
    # RT130 data looks to be all big-endian (logpeek assumes this, and it has
    # been working pretty well), so we don't have to do any endianness check.

    packets = []

    with open(file_name, 'rb') as rt130_file:
        # Each packet is exactly 1024 bytes, so we can rely on that to know
        # when we have finished reading.
        for i in range(os.path.getsize(file_name) // 1024):
            packet = rt130_file.read(1024)
            packet_header = get_rt130_packet_header(packet)

            waveform_handlers: Dict[str, Callable] = {
                'EH': read_eh_et_packet,
                'ET': read_eh_et_packet,
                'DT': read_dt_packet,
            }

            soh_handlers: Dict[str, Callable] = dict.fromkeys(
                ['AD', 'CD', 'DS', 'FD', 'OM', 'SC', 'SH'],
                read_soh_packet
            )

            packet_handlers = {
                **waveform_handlers, **soh_handlers
            }

            packet_handler = packet_handlers.get(
                packet_header.packet_type, packet_reader_placeholder
            )
            return_val = packet_handler(packet, unpacker)
            if packet_header.packet_type == 'DT':
                packet_type = DTPacket
            elif packet_header.packet_type in ['EH', 'ET']:
                packet_type = EHETPacket
            else:
                packet_type = SOHPacket

            extended_header, data = return_val
            current_packet = packet_type(packet_header, extended_header, data)
            packets.append(current_packet)

    return packets


def convert_packet_to_obspy_format(packet: Union[EHETPacket, DTPacket,
                                                 SOHPacket],
                                   unpacker: Unpacker) -> Tuple:
    """
    Convert an RT130 packet into a numpy array of type PACKET_FINAL_DTYPE
    :param packet: an RT130 packet.
    :param unpacker: the decoder used to decode the data.
    :return: a tuple that can be converted into an object of type
        PACKET_FINAL_DTYPE that contains the data stored in packet.
    """
    # We want to convert the packet to a tuple. In order to make it easier to
    # maintain, we first convert the packet to a dictionary. Then, we grab the
    # values of the dictionary as tuple to get the final result.
    converted_packet = {}
    converted_packet['packet_type'] = packet.header.packet_type
    converted_packet['experiment_number'] = packet.header.experiment_number
    # Obspy only stores the last two digits of the year.
    converted_packet['year'] = packet.header.time.year % 100
    converted_packet['unit_id'] = packet.header.unit_id
    converted_packet['time'] = packet.header.time.ns
    converted_packet['byte_count'] = packet.header.byte_count
    converted_packet['packet_sequence'] = packet.header.packet_sequence
    converted_packet['event_number'] = packet.extended_header.event_number
    converted_packet[
        'data_stream_number'] = packet.extended_header.data_stream_number
    converted_packet['channel_number'] = packet.extended_header.channel_number
    converted_packet[
        'number_of_samples'] = packet.extended_header.number_of_samples
    converted_packet['flags'] = packet.extended_header.flags
    converted_packet['data_format'] = packet.extended_header.data_format

    if converted_packet['packet_type'] == 'DT':
        # Obspy stores the data as list of 1-byte integers. We store the
        # data as an arbitrary length integer, so we need to do some
        # conversion. To make encoding and decoding the data point easier, we
        # store it in 4 bytes no matter what the data format is. This only
        # has an effect on data with format 16. Thanks to a quirk with
        # 2-complement binary encoding, however, this does not cause any
        # problem.
        data_size = 4
        format_char = 'B'
        converted_packet['payload'] = numpy.empty(1000, np.uint8)
        packet_data = list(unpacker.unpack(
            f'{data_size}{format_char}',
            packet.data.to_bytes(data_size, 'big', signed=True)
        ))
        converted_packet['payload'][:4] = packet_data
    elif converted_packet['packet_type'] in ['EH', 'ET']:
        eh_et_payload_size = eh_et_payload_end_in_packet - 24
        converted_packet['payload'] = numpy.empty(1000, np.uint8)
        packet_data = numpy.frombuffer(packet.data, np.uint8)
        converted_packet['payload'][:eh_et_payload_size] = packet_data
    else:
        converted_packet['payload'] = numpy.frombuffer(packet.data, np.uint8)
    return tuple(converted_packet.values())
